-- Aligner: A Tojaso FowlPlay Project
-- Original Author: Tomber, all rights reserved
-- Current Author: Casstiel 
-- 		Tomber has given Casstiel permission to distribute and support Aligner during his hiatus from Wildstar. 

require "GameLib"
require "Window"

Aligner = {}
local spacing = 20 -- space between lines, both horizontally and vertically
local opacity = 0.5 -- opacity of the grid lines
local hidden = false -- set to true when hiding the grid lines
local lineColor = CColor.new(0, 1, 1, 1) -- cyan is the default color for most lines
local centerColor = CColor.new(1, 0, 0, 1) -- red is the default color for the center lines
local horizontals, verticals = {}, {} -- tables of cached horizonal and vertical lines
local hcount, vcount = 0, 0 -- number of currently used horizontal and vertical lines

-- Addon initialization and registration
function Aligner:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

function Aligner:Init()
    Apollo.RegisterAddon(self)
end

-- Allocate horizontal and vertical lines as needed
function Aligner:AllocateLines(h, v)
	assert(type(h) == "number" and h > 0 and type(v) == "number" and v > 0, "Aligner: error allocating lines")
	local nh, nv = #horizontals, #verticals
	if h > nh then
		for i = nh + 1, h do horizontals[i] = Apollo.LoadForm(self.xmlDoc, "HorizontalLine", "FixedHudStratum", self) end
	end
	if v > nv then
		for i = nv + 1, v do verticals[i] = Apollo.LoadForm(self.xmlDoc, "VerticalLine", "FixedHudStratum", self) end
	end
end

-- Draw a horizontal line of width w with color c at the given vertical offset y
function Aligner:DrawHorizontal(w, y, c)
	hcount = hcount + 1
	local line = horizontals[hcount]
	assert(line, "Aligner: not enough horizontals allocated")
	line:SetAnchorOffsets(0, y - 0.5, w, y + 0.5); line:SetBGColor(c); line:SetOpacity(opacity); line:Show(true)
end

-- Draw a vertical line of height h with color c at the given horizontal offset x
function Aligner:DrawVertical(x, h, c)
	vcount = vcount + 1
	local line = verticals[vcount]
	assert(line, "Aligner: not enough verticals allocated")
	line:SetAnchorOffsets(x - 0.5, 0, x + 0.5, h); line:SetBGColor(c); line:SetOpacity(opacity); line:Show(true)
end

-- Update values in controls from current values
function Aligner:UpdateControls()
	local ap = Aligner
	ap.lineColor:SetBGColor(lineColor)
	ap.centerColor:SetBGColor(centerColor)
	local op = math.floor((opacity * 100) + 0.5); ap.opacity:SetValue(op); ap.opacityCaption:SetText(tostring(op))
	local sp = math.floor(spacing + 0.5); ap.spacing:SetValue(sp); ap.spacingCaption:SetText(tostring(sp))
	if hidden then -- hide all grid lines
		for _, h in pairs(horizontals) do h:Show(false) end
		for _, v in pairs(verticals) do v:Show(false) end
		hcount = 0; vcount = 0
		ap.toggle:SetText("Show")
	else -- show grid lines
		local d = Apollo.GetDisplaySize()
		local w, h = d.nWidth, d.nHeight
		self:AllocateLines(math.floor((h / spacing) + 1), math.floor((w / spacing) + 1)) -- make sure that there are enough lines allocated
		hcount = 0; vcount = 0 -- these are incremented as a side effect of the draw functions
		local xc, yc = w / 2, h / 2 -- coordinates of the center of the display
		self:DrawHorizontal(w, yc, centerColor); self:DrawVertical(xc, h, centerColor) -- draw the center horizontal and vertical lines
		for k = 1, math.floor(h / (2 * spacing)) do -- figure out how many pairs of horizontal lines to draw
			self:DrawHorizontal(w, yc - (k * spacing), lineColor); self:DrawHorizontal(w, yc + (k * spacing), lineColor) -- draw pair of horizontals
		end
		for k = 1, math.floor(w / (2 * spacing)) do -- figure out how many pairs of vertical lines to draw
			self:DrawVertical(xc - (k * spacing), h, lineColor); self:DrawVertical(xc + (k * spacing), h, lineColor) -- draw pair of verticals
		end
		for k = hcount + 1, #horizontals do horizontals[k]:Show(false) end -- hide any extra horizontal and vertical lines
		for k = vcount + 1, #verticals do verticals[k]:Show(false) end
		ap.toggle:SetText("Hide")
	end
end

-- At load time, plug in our custom images
function Aligner:OnLoad()
	self.wndMain = nil
	self.xmlDoc = XmlDoc.CreateFromFile("Aligner.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- Aligner OnDocLoaded
-----------------------------------------------------------------------------------------------
function Aligner:OnDocLoaded()
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "AlignerForm", nil, self)
	self.wndMain:Show(false) -- hide the options window by default

	Apollo.RegisterSlashCommand("aligner", "OnAlignerOn", self)
	Apollo.RegisterSlashCommand("align", "OnAlignerOn", self)
	Apollo.RegisterEventHandler("ResolutionChanged", "OnResolutionChanged", self)

	local ap = Aligner
	ap.toggle = self.wndMain:FindChild("ToggleButton")
	ap.lineColor = self.wndMain:FindChild("ColorSwatch")
	ap.centerColor = self.wndMain:FindChild("CenterSwatch")
	local spacingBase = self.wndMain:FindChild("SpacingBase")
	if spacingBase then ap.spacing = spacingBase:FindChild("SliderBar"); ap.spacingCaption = spacingBase:FindChild("Amount") end
	local opacityBase = self.wndMain:FindChild("OpacityBase")
	if opacityBase then ap.opacity = opacityBase:FindChild("SliderBar"); ap.opacityCaption = opacityBase:FindChild("Amount") end
	assert(ap.toggle and ap.lineColor and ap.centerColor and ap.spacing and ap.spacingCaption and ap.opacity and ap.opacityCaption,
		"Aligner: form initialization error")
end

-- At update time, need to make sure colors are correct if they are being edited in ColorPicker
local function OnColorUpdate()
	if not hidden and hcount > 0 and vcount > 0 then
		for k = 1, hcount do horizontals[k]:SetBGColor(k == 1 and centerColor or lineColor) end -- update colors of horizontal lines
		for k = 1, vcount do verticals[k]:SetBGColor(k == 1 and centerColor or lineColor) end -- update colors of vertical lines
	end
	local ap = Aligner
	if ap.lineColor and ap.centerColor then
		ap.lineColor:SetBGColor(lineColor) -- update color swatches dynamically in case being edited
		ap.centerColor:SetBGColor(centerColor)
	end
end

-- If display resolution changes then redraw
function Aligner:OnResolutionChanged()
	if not hidden and hcount > 0 and vcount > 0 then
		hidden = true; UpdateControls(); hidden = false; UpdateControls() -- hide and then redraw everything
	end
end

-- Respond to registered slash commands
function Aligner:OnAlignerOn()
	self:UpdateControls() -- update values in controls
	self.wndMain:Show(true) -- show the window
	self.wndMain:ToFront() -- bring it to the front
end

-- Respond to OK button click
function Aligner:OnOK()
	self.wndMain:Show(false) -- hide the options window but don't automatically hide the lines
end

-- Respond to Show/Hide button click to toggle line display
function Aligner:OnToggle()
	hidden = not hidden
	self:UpdateControls() -- update current values
end

-- Respond to spacing slider changes
function Aligner:OnSpacingChanging(wndHandler, wndControl, x)
	if x and x > 1 then spacing = x end -- enforce minimum spacing
	self:UpdateControls() -- update current values
end

-- Respond to opacity slider changes
function Aligner:OnOpacityChanging(wndHandler, wndControl, x)
	opacity = x / 100
	self:UpdateControls() -- update current values
end

-- Respond to color swatch click by launching color picker
function Aligner:OnColorSwatch()
	if ColorPicker then ColorPicker.AdjustCColor(lineColor, false, OnColorUpdate) end
	self:UpdateControls() -- update current values
end

-- Respond to center swatch click by launching color picker
function Aligner:OnCenterSwatch()
	if ColorPicker then ColorPicker.AdjustCColor(centerColor, false, OnColorUpdate) end
	self:UpdateControls() -- update current values
end

-- Convert a color object into a hex-encoded string using format "rrggbb"
local function Convert_CColor_To_String(c)
	return string.format("%02x%02x%02x", math.floor(c.r * 255 + 0.5), math.floor(c.g * 255 + 0.5), math.floor(c.b * 255 + 0.5))
end

-- Convert a hex-encoded string into a color object, string format is assumed to be "rrggbb"
local function Convert_String_To_CColor(hex)
	local r, g, b = 0, 0, 0 -- invalid strings will result in these values being returned
	local n = tonumber(hex, 16)
	if n then r = math.floor(n / 65536); g = math.floor(n / 256) % 256; b = n % 256 end
	return CColor.new(r / 255, g / 255, b / 255, 1)
end

-- Save current settings for next time
function Aligner:OnSave(level)
	local profile = nil
	if level == GameLib.CodeEnumAddonSaveLevel.General then
		profile = {
			spacing = math.floor(spacing + 0.5),
			opacity = math.floor((opacity * 100) + 0.5),
			lineColor = Convert_CColor_To_String(lineColor),
			centerColor = Convert_CColor_To_String(centerColor),
		}
	end
	return profile
end

-- Restore settings from profile created during previous use the addon on any character
function Aligner:OnRestore(level, profile)
	if profile and level == GameLib.CodeEnumAddonSaveLevel.General then
		if profile.spacing then spacing = profile.spacing end
		if profile.opacity then opacity = profile.opacity / 100 end
		if profile.lineColor then lineColor = Convert_String_To_CColor(profile.lineColor) end
		if profile.centerColor then centerColor = Convert_String_To_CColor(profile.centerColor) end
	end
end

-- Finally, just need to instantiate the addon...
local AlignerInst = Aligner:new()
AlignerInst:Init()
